use cpal::Sample;
use tracing::{trace};

const TAU: f64 = 2.0 * std::f64::consts::PI;

pub struct TwoStateCarrier {
    pub sample_rate: f64,
    pub freq_low: f64,
    pub freq_high: f64,
    phase_low: f64,
    phase_high: f64,
}

type PcmSample = i64;

pub trait Carrier {
    fn byte_into_fsk_samples(&mut self, byte: &u8) -> Vec<PcmSample>;
}

impl Carrier for TwoStateCarrier {
    fn byte_into_fsk_samples(&mut self, byte: &u8) -> Vec<PcmSample> {
        let mut bits: Vec<u8> = vec![];
        let mut out: Vec<i64> = vec![];

        // MSb first
        bits.push((byte     ) & 1);
        bits.push((byte >> 1) & 1);
        bits.push((byte >> 2) & 1);
        bits.push((byte >> 3) & 1);
        bits.push((byte >> 4) & 1);
        bits.push((byte >> 5) & 1);
        bits.push((byte >> 6) & 1);
        bits.push((byte >> 7) & 1);

        for (i, bit) in bits.into_iter().enumerate() {
            if bit == 0 {
                trace!("low bit {} in fsk code", i);
                while (self.phase_low) < self.sample_rate / self.freq_low {
                    trace!("low bit gen sin");
                    out.push((self.phase_low * TAU * self.freq_low / self.sample_rate).sin().to_sample::<i64>());
                    self.phase_low += 1.0;
                }
                self.phase_low = self.phase_low % self.sample_rate;
            } else if bit == 1 {
                trace!("high bit {} in fsk code", i);
                while (self.phase_high) < self.sample_rate / self.freq_high {
                    trace!("high bit gen sin");
                    out.push((self.phase_high * TAU * self.freq_low / self.sample_rate).sin().to_sample::<i64>());
                    self.phase_high += 1.0;
                }
                self.phase_high = self.phase_high % self.sample_rate;
            }
        }
        trace!("FSK calculations turned up {:?}", &out);
        out
    }
}

impl TwoStateCarrier {
    pub fn new(freq_low: u32, freq_high: u32, sample_rate: u32) -> Self {
        TwoStateCarrier {
            freq_low: freq_low.into(),
            freq_high: freq_high.into(),
            sample_rate: sample_rate.into(),
            phase_low: 0_f64,
            phase_high: 0_f64,
        }
    }
    fn freq_low(mut self, freq_low: u32) {
        self.freq_low = freq_low.into();
    }
    fn freq_high(mut self, freq_high: u32) {
        self.freq_high = freq_high.into();
    }
    fn phase_low(mut self, phase_low: u32) {
        self.phase_low = phase_low.into();
    }
    fn phase_high(mut self, phase_high: u32) {
        self.phase_high = phase_high.into();
    }
}
